
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/time.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <fcntl.h>
#include <sys/ioctl.h>


enum {
//	_xinfreq = 20_000_000,
	_xtlfreq = 20_000_000,
	_clkfreq = 200_000_000,

	HEAPSIZE = 4400,

	SD_BASE_PIN = 16,
	PIN_DAT    = SD_BASE_PIN | 3<<6,
	PIN_DAT0   = SD_BASE_PIN+0,    // MISO
	PIN_DAT1   = SD_BASE_PIN+1,
	PIN_DAT2   = SD_BASE_PIN+2,
	PIN_DAT3   = SD_BASE_PIN+3,    // CS
	PIN_CMD    = SD_BASE_PIN+4,    // MOSI
	PIN_CLK    = SD_BASE_PIN+5,    // SCLK
	PIN_LED    = SD_BASE_PIN+6,  // output to red LED
	PIN_POW    = SD_BASE_PIN+7,  // Card-Detect input, and power switch, and green LED
/*
	SD_BASE_PIN = 40,
	PIN_DAT0   = SD_BASE_PIN+4,    // MISO
	PIN_CMD    = SD_BASE_PIN+2,    // MOSI
	PIN_CLK    = SD_BASE_PIN+3,    // SCLK
*/
	PIN_DO    = PIN_DAT0,
	PIN_DI    = PIN_CMD,
	PIN_CS    = PIN_DAT0+3,
};



static uint32_t  randfill( uint32_t *addr, size_t size, uint32_t state )    // "size" is number of longwords
{
    __asm volatile {    // "const" enforces XIP, "volatile" enforces Fcache
		wrfast	#0, addr
		rep	@.rend, size
		xoro32	state
		mov	pb, 0-0
		wflong	pb
.rend
    }

    return state;    // for potential continuation of psuedo random sequence
}



static size_t  randcmp( const uint32_t *addr, size_t size, uint32_t state )    // "size" is number of longwords
{
    uint32_t  count = 0;

    __asm volatile {    // "const" enforces XIP, "volatile" enforces Fcache
		rdfast	#0, addr
		rep	@.rend, size
		rflong	pb
		xoro32	state
		cmp	pb, 0-0   wz
	if_z	add	count, #1
.rend
    }

    return  count;    // pass longword count
}



static void  tester( const char *filename, size_t kbytes, size_t repeats )
{
    FILE  *fh;
    size_t  count, i;
    uint32_t  tmr, seed, rstate, *buff;
    const size_t  bytes = kbytes * 1024;

    buff = __builtin_alloca(bytes);    // auto-frees upon return from tester()
    if( !buff ) {    // auto-frees upon return from tester()
        printf(" malloc() failed!\n");
        exit(1);
    }

//-----------------------
// Write file
//-----------------------
    seed = _rnd();
    rstate = seed;
    rstate = randfill(buff, kbytes * 256, rstate);

    _seterror(0);
    fh = fopen(filename, "wb");
    if( !fh ) {
        printf(" fopen() for writing failed!   errno = %d: %s\n", errno, strerror(errno));
        exit(3);
    }
    printf(" Buffer = %d kB, ", kbytes);

    count = 0;
    tmr = _getus();
    for( i = 0; i < repeats; i++ ) {
//        rstate = randfill(buff, kbytes * 256, rstate);
        count += fwrite(buff, 1, bytes, fh) / 1024;
    }
    fclose(fh);
    tmr = _getus() - tmr;
    printf(" Written %d kB at %d kB/s, ", count,
            _muldiv64(count, 1_000_000, tmr));
    if( count != kbytes * repeats ) {
        printf(" Error: File not complete!   errno = %d: %s\n", errno, strerror(errno));
        exit(4);
    }

//-----------------------
// Read back file
//-----------------------
    _seterror(0);
    fh = fopen(filename, "rb");
    if( !fh ) {
        printf(" fopen() for readback failed!   errno = %d: %s\n", errno, strerror(errno));
        exit(5);
    }
    count = kbytes * repeats;
    for( i = 0; i < repeats; i++ ) {
//        rstate = seed;
//        seed = randfill(buff, kbytes * 256, rstate);
        if( bytes != fread(buff, 1, bytes, fh) )
            break;
        count -= randcmp(buff, kbytes * 256, seed) / 256;
//        count -= randcmp(buff, kbytes * 256, rstate) / 256;
    }
    if( count )
        printf(" Mis-match! ");
    else
        printf(" Verified, ");

    rewind(fh);    // read the file again, this time purely for speed check
    tmr = _getus();
    for( i = 0; i < repeats; i++ )
        count += fread(buff, 1, bytes, fh) / 1024;
    fclose(fh);
    tmr = _getus() - tmr;
    printf(" Read %d kB at %d kB/s\n", count,
           _muldiv64(count, 1_000_000, tmr));
}



static void  mountsd( void )
{
    int  rc, clkdiv;

    umount("/sd");
//    mount("/sd", _vfs_open_sdcard());

    _seterror(0);
//    FILE *raw_device = _sdmm_open(PIN_CLK, PIN_CS, PIN_DI, PIN_DO);
    FILE *raw_device = _sdsd_open(PIN_CLK, PIN_CMD, PIN_DAT0, PIN_POW, PIN_LED);
//    FILE *raw_device = _sdsd_open(PIN_CLK, PIN_CMD, PIN_DAT0, -1, -1);
    if(!raw_device) {
        printf(" device open failed!   errno = %d: %s\n", errno, strerror(errno));
        exit(1);
    }
//    rc = raw_device->ioctl(raw_device, 71, buff);    // get CLK_DIV value
//    printf(" ioctl 71 CLKDIV=%d rc=%d errno=%d:%s\n", buff[0], rc, errno, strerror(errno));
    clkdiv = 0;
//    rc = raw_device->ioctl(raw_device, 70, &clkdiv);    // disable read-block CRC processing
    clkdiv = 2;
//    rc = raw_device->ioctl(raw_device, 72, &clkdiv);    // set CLK_DIV value

    mount("/sd", _vfs_open_fat_handle(raw_device));

}



static void  shutdown( void )
{
    umount("/sd");
    puts("\nexit\n");  // blank line for re-run separation
    _waitms(500);
    _clkset(1, 20_000);  // cool running
}


//===================================================================================

void  main( void )
{
    int  i;

    atexit(shutdown);
    printf("   clkfreq = %d   clkmode = 0x%x\n", _clockfreq(), _clockmode());

    printf("  Compiled with FlexC v%s\n", __VERSION__);

    mountsd();

/*
    puts("");
    for( i = 0; i < 4; i++ )
        tester( "/sd/speed1.bin", 2, 1024 );    // 1024 x 2 KB = 2 MB
*/
    puts("");
    for( i = 0; i < 4; i++ )
        tester( "/sd/speed2.bin", 4, 512 );    // 512 x 4 KB = 2 MB
/*
    puts("");
    for( i = 0; i < 4; i++ )
        tester( "/sd/speed3.bin", 8, 512 );    // 512 x 8 KB = 4 MB
    puts("");
    for( i = 0; i < 4; i++ )
        tester( "/sd/speed4.bin", 16, 256 );    // 256 x 16 KB = 4 MB
    puts("");
    for( i = 0; i < 4; i++ )
        tester( "/sd/speed5.bin", 32, 256 );    // 256 x 32 KB = 8 MB
*/
    puts("");
    for( i = 0; i < 4; i++ )
        tester( "/sd/speed6.bin", 64, 128 );    // 128 x 64 KB = 8 MB
/*
    puts("");
    for( i = 0; i < 4; i++ )
        tester( "/sd/speed7.bin", 128, 128 );    // 128 x 128 KB = 16 MB
    puts("");
    for( i = 0; i < 4; i++ )
        tester( "/sd/speed8.bin", 256, 64 );    // 64 x 256 KB = 16 MB

    for( i = 0; i < 3; i++ )
        tester( "/sd/speed1.bin", 2, 32 );    // 32 x 2 KB = 64 kB
    puts("");
    for( i = 0; i < 3; i++ )
        tester( "/sd/speed2.bin", 4, 32 );    // 32 x 4 KB = 128 kB
    puts("");
    for( i = 0; i < 3; i++ )
        tester( "/sd/speed3.bin", 8, 32 );    // 32 x 8 KB = 256 kB
    puts("");
    for( i = 0; i < 4; i++ )
        tester( "/sd/speed4.bin", 16, 32 );    // 32 x 16 KB = 512 kB
    puts("");
    for( i = 0; i < 4; i++ )
        tester( "/sd/speed5.bin", 32, 32 );    // 32 x 32 KB = 1 MB
    puts("");
    for( i = 0; i < 4; i++ )
        tester( "/sd/speed6.bin", 64, 32 );    // 32 x 64 KB = 2 MB

    puts("");
    for( i = 0; i < 4; i++ )
        tester( "/sd/speed7.bin", 128, 2 );    // 2 x 128 KB = 256 kB

    puts("");
    for( i = 0; i < 4; i++ )
        tester( "/sd/speed8.bin", 256, 32 );    // 32 x 256 KB = 8 MB
    puts("");
    for( i = 0; i < 4; i++ )
        tester( "/sd/speed8.bin", 440, 32 );    // 32 x 448 KB = 14 MB
*/
    exit(0);
}
